package com.xkrs.service.impl;

import com.xkrs.common.encapsulation.PromptMessageEnum;
import com.xkrs.dao.CountyCodeWeiXinDao;
import com.xkrs.dao.FirePointDao;
import com.xkrs.dao.StreetDao;
import com.xkrs.model.entity.CountyCodeWeiXinEntity;
import com.xkrs.model.entity.FirePointEntity;
import com.xkrs.model.qo.FirePointQo;
import com.xkrs.model.vo.GaoDeIgGeocodeVo;
import com.xkrs.service.FirePointService;
import com.xkrs.service.StreetService;
import com.xkrs.straw.dao.FirePointOrdinaryDao;
import com.xkrs.straw.dao.SysUserDao;
import com.xkrs.straw.model.bean.FirePointQueryResultBean;
import com.xkrs.straw.model.entity.FirePointOrdinaryEntity;
import com.xkrs.straw.model.entity.SysUserEntity;
import com.xkrs.straw.model.helper.PushHelper;
import com.xkrs.straw.model.helper.SMSHelper;
import com.xkrs.straw.model.vo.AllFirePointVo;
import com.xkrs.straw.utils.FirePointConvertUtils;
import com.xkrs.straw.utils.FirePointQueryManager;
import com.xkrs.utils.*;
import org.apache.hc.core5.util.TextUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.cache.annotation.CacheConfig;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import javax.persistence.criteria.Predicate;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.text.DecimalFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.*;

import static com.xkrs.common.encapsulation.OutputEncapsulation.outputEncapsulationObject;
import static com.xkrs.utils.DateTimeUtil.dateTimeToString;
import static com.xkrs.utils.XlsxUtil.downloadXlsx;

/**
 * @author XinYi Song
 */
@CacheConfig(cacheNames = "FirePointServiceCache")
@Service
public class FirePointServiceImpl implements FirePointService {

    public static Logger log = LoggerFactory.getLogger(FirePointServiceImpl.class);

    private final Locale locale = LocaleContextHolder.getLocale();

    @Resource
    private FirePointDao firePointDao;

    @Resource
    private SysUserDao sysUserDao;

    @Resource
    private StreetDao streetDao;

    @Resource
    private CountyCodeWeiXinDao countyCodeWeiXinDao;

    @Resource
    private FirePointQueryHelper firePointQueryHelper;

    @Resource
    private StreetService streetService;

    @Resource
    private SMSHelper smsHelper;

    @Resource
    private PushHelper pushHelper;

    @Resource
    private FirePointQueryManager firePointQueryManager;

    @Resource
    private FirePointOrdinaryDao firePointOrdinaryDao;

    /**
     * 添加火点信息
     */
    @CacheEvict(value = "FirePointServiceCache", allEntries = true)
    @Override
    public boolean insertFirePoint(FirePointQo firePointQo) {
        if ("水体".equals(firePointQo.getLandtype())) {
            return false;
        }
        if ("Himawari 8".equals(firePointQo.getSatelliteType())) {
            if ("N".equals(firePointQo.getConfidence()) || "L".equals(firePointQo.getConfidence())) {
                return false;
            }
        }
        FirePointEntity firePointEntity = new FirePointEntity();
        firePointEntity.setFireCode(firePointQo.getFireCode());
        firePointEntity.setCountyCode(firePointQo.getCountyCode().toString());
        firePointEntity.setCountyName(firePointQo.getCountyName());
        firePointEntity.setSatelliteTime(DateTimeUtil.timeMillisToString(firePointQo.getSatelliteTimeTs().longValue()));
        firePointEntity.setLongitude(firePointQo.getLongitude());
        firePointEntity.setLatitude(firePointQo.getLatitude());
        firePointEntity.setSatelliteType(firePointQo.getSatelliteType());
        firePointEntity.setLandType(firePointQo.getLandtype());
        firePointEntity.setConfidence(firePointQo.getConfidence());
        firePointEntity.setAddTime(dateTimeToString(LocalDateTime.now()));
        firePointEntity.setFireType("0");
        firePointEntity.setFireImage(firePointQo.getFireImage());
        firePointEntity.setSatelliteImage(firePointQo.getSatelliteImage());
        firePointEntity.setAuditFireType("0");//初始化火点为未审核状态
        bindAddress(firePointEntity);
        if (checkDuplicatedData(firePointEntity)) {
            log.info("-------该火点和历史火点重复，不入库");
            return false;
        }
        log.info("-------发现新火点");
        firePointDao.save(firePointEntity);
        //发送消息通知
        sendBroadcast(firePointEntity);
        return true;
    }

    /**
     * 检查重复数据
     *
     * @param firePointEntity 待入库实体类
     * @return true-是重复数据，拒绝入库；false-不是重复数据，可以入库
     */
    private boolean checkDuplicatedData(FirePointEntity firePointEntity) {
        List<FirePointEntity> duplicatedDataList = firePointDao.findDuplicatedData(firePointEntity.getLongitude(), firePointEntity.getLatitude(), firePointEntity.getStreetCode(), firePointEntity.getSatelliteTime(), firePointEntity.getSatelliteType(), firePointEntity.getLandType());
        return duplicatedDataList != null && duplicatedDataList.size() > 0;
    }

    /**
     * 为实体类绑定地址、编号、名称
     */
    private void bindAddress(FirePointEntity firePointEntity) {
        List<String> locationList = new ArrayList<>();
        locationList.add(firePointEntity.getLongitude() + "," + firePointEntity.getLatitude());
        GaoDeIgGeocodeVo geocode = GaoDeApiUtil.geocode(locationList);
        GaoDeIgGeocodeVo.Regeocode reGeoCode = geocode.getRegeocodes().get(0);
        String formattedAddress = reGeoCode.getFormatted_address();
        firePointEntity.setFirePointAddress(formattedAddress);
        GaoDeIgGeocodeVo.AddressComponent addressComponent = reGeoCode.getAddressComponent().get(0);
        firePointEntity.setCountyCode(addressComponent.getAdcode());
        firePointEntity.setCountyName(addressComponent.getDistrict());
        String townCode = addressComponent.getTowncode();
        String fixedTownCode = townCode.length() > 9 ? townCode.substring(0, 9) : townCode;
        firePointEntity.setStreetCode(fixedTownCode);
        firePointEntity.setStreetName(addressComponent.getTownship());
    }

    private void sendBroadcast(FirePointEntity firePointEntity) {
        //发送微信群聊消息
        sendWeChatGroupMessage(firePointEntity);

        if ("耕地".equals(firePointEntity.getLandType())) {
            //过滤短信通知收件人
            List<SysUserEntity> sysUserList = sysUserDao.findAll();
            List<SysUserEntity> smsReceiverList = FirePointCodeUtils.filterSmsReceiver(firePointEntity.getStreetCode(), sysUserList);
//            //推送
//            try {
//                List<String> userAccountList = pushHelper.obtainUserAccountList(smsReceiverList);
//                pushHelper.dispatchPushMessage(userAccountList, firePointEntity);
//            } catch (Exception e) {
//                e.printStackTrace();
//            }
            //分发短信通知
            smsHelper.dispatchSMSMessage(smsReceiverList, new String[]{firePointEntity.getFirePointAddress()}, new FirePointEntity[]{firePointEntity});
        }
    }

    /**
     * 发送微信消息
     *
     * @param firePointEntity
     */
    private void sendWeChatGroupMessage(FirePointEntity firePointEntity) {
        List<CountyCodeWeiXinEntity> countyCodeWeiXinList = countyCodeWeiXinDao.findAll();
        if (countyCodeWeiXinList.isEmpty()) {
            return;
        }
        List<String> weixinIdList = new ArrayList<>();
        for (CountyCodeWeiXinEntity countyCodeWeiXin : countyCodeWeiXinList) {
            try {
                if ((countyCodeWeiXin.getSendState() != null) && (countyCodeWeiXin.getSendState() != 0)) {
                    String formatCutCode = FirePointCodeUtils.getFormatCutCode(countyCodeWeiXin.getCountyCode());
                    if (firePointEntity.getStreetCode().startsWith(formatCutCode)) {
                        weixinIdList.add(countyCodeWeiXin.getWeixinId());
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (weixinIdList.size() > 0) {
            StringBuilder stringBuilder = new StringBuilder("发送微信消息通知：\n");
            String messageContent = getMessageContent(firePointEntity);
            for (String weixinId : weixinIdList) {
                stringBuilder.append(weixinId).append("\n");
                try {
                    WDWxSendMsgUtil.sendMsg(weixinId, messageContent, 0);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            stringBuilder.append("火情信息：").append(messageContent);
            log.info(stringBuilder.toString());
        }
    }

    private String getMessageContent(FirePointEntity firePointEntity) {
        java.text.DecimalFormat decimalFormat = new DecimalFormat("#.000000");
        String satelliteTime = firePointEntity.getSatelliteTime();
        String formatLongitude = decimalFormat.format(firePointEntity.getLongitude());
        String formatLatitude = decimalFormat.format(firePointEntity.getLatitude());
        String countyName = firePointEntity.getCountyName();
        String streetName = firePointEntity.getStreetName();
        String landType = firePointEntity.getLandType();
        return firePointEntity.getSatelliteType() + "发现1个火点。\n卫星时间：" + satelliteTime + "；\nlongitude：" + formatLongitude + "；\nlatitude：" + formatLatitude + "；\ncountyName：" + countyName + "；\nstreetName：" + streetName + "；\nlandType：" + landType;
    }

    /**
     * 下载VIP用户火点数据
     */
    @Override
    public String downloadVipUserFilePoint(HttpServletRequest request, HttpServletResponse response) {
        LocalDate now = LocalDate.now();
        List<Map<String, String>> list = firePointDao.downloadFirePointByVip(dateTimeToString(LocalDateTime.of(now.minusDays(1), LocalTime.MIN)));
        if (list.size() == 0) {
            return outputEncapsulationObject(PromptMessageEnum.SUCCESS, "无数据!", locale);
        }
        int year = now.getYear();
        int month = now.getMonth().getValue();
        int day = now.getDayOfMonth();
        String fileName = String.format("%s_%s_%s_VIP用户火点信息列表.xlsx", year, month, day);
        response.setContentType("application/octet-stream; charset=UTF8");
        response.setHeader("Content-Disposition", "attachment; filename=" + URLEncoder.encode(fileName, StandardCharsets.UTF_8));
        response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
        try {
            downloadXlsx(list, response.getOutputStream());
        } catch (IOException e) {
            log.error("download vip user's fire point fail: {}", e.getMessage());
        }
        return outputEncapsulationObject(PromptMessageEnum.SUCCESS, "ok", locale);
    }

    @Override
    public String downloadFirePoint(HttpServletRequest request, HttpServletResponse response, String ids) {

        List<String> idList = ListUtils.toStringList(ids, ",");
        if (idList.isEmpty()) {
            return outputEncapsulationObject(PromptMessageEnum.SUCCESS, "无数据!", locale);
        }
        List<Map<String, String>> list = new ArrayList<>();
        List<FirePointEntity> sourceList = firePointDao.findAllById(ListUtils.stringListToLongList(idList));
        for (FirePointEntity source : sourceList) {
            Map<String, String> record = new HashMap<>();
            record.put("所在区县", source.getCountyName());
            record.put("所在街道", source.getStreetName());
            record.put("发现时间", source.getSatelliteTime());
            record.put("地物类型", source.getLandType());
            record.put("经度", String.valueOf(source.getLongitude()));
            record.put("纬度", String.valueOf(source.getLatitude()));
            record.put("详细地址", source.getFirePointAddress());
            list.add(record);
        }
        LocalDate now = LocalDate.now();
        String fileName = String.format("%s_%s_%s_火点信息列表.xlsx", now.getYear(), now.getMonth().getValue(), now.getDayOfMonth());
        response.setContentType("application/octet-stream; charset=UTF8");
        response.setHeader("Content-Disposition", "attachment; filename=" + URLEncoder.encode(fileName, StandardCharsets.UTF_8));
        response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
        try {
            downloadXlsx(list, response.getOutputStream());
        } catch (IOException e) {
            log.error("download fire point fail: {}", e.getMessage());
        }
        return outputEncapsulationObject(PromptMessageEnum.SUCCESS, "ok", locale);
    }

    /**
     * Complete
     */
    @Override
    public String queryFirePoint(String token, String code, String startTime, String endTime, String satelliteType, String landType) {
        //验证token
        String userName = TokenUtil.obtainUserNameByToken(token);
        if (TextUtils.isEmpty(userName)) {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, "账号不存在或错误", locale);
        }
        SysUserEntity sysUserEntity = sysUserDao.selectByUserName(userName);
        if (sysUserEntity == null) {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, "账号不存在或错误", locale);
        }

        FirePointQueryResultBean firePointQueryResult = firePointQueryManager.queryFirePointBelongToUser(sysUserEntity, code, startTime, endTime, satelliteType, landType);
        if (firePointQueryResult.isSuccess()) {
            return outputEncapsulationObject(PromptMessageEnum.SUCCESS, firePointQueryResult.getList(), locale);
        } else {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, firePointQueryResult.getErrorMessage(), locale);
        }
    }

    /**
     * Complete
     */
//    @Cacheable(keyGenerator = "keyGenerator", unless = "#result == null")
    @Override
    public String selectTodayFirePoint(String token, String countyCode) {

        //验证token
        String userName = TokenUtil.obtainUserNameByToken(token);
        if (TextUtils.isEmpty(userName)) {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, "账号不存在或错误", locale);
        }
        SysUserEntity sysUserEntity = sysUserDao.selectByUserName(userName);
        if (sysUserEntity == null) {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, "账号不存在或错误", locale);
        }

        LocalDateTime startTime1 = LocalDateTime.now();
        LocalDateTime startTime2 = LocalDateTime.of(startTime1.getYear(), startTime1.getMonth(), startTime1.getDayOfMonth(), 0, 0, 0, 0);
        String startTimeString = DateTimeUtils.localDateTimeToString(startTime2);

        LocalDateTime endTime1 = LocalDateTime.now().plusDays(1);
        LocalDateTime endTime2 = LocalDateTime.of(endTime1.getYear(), endTime1.getMonth(), endTime1.getDayOfMonth(), 0, 0, 0, 0);
        String endTimeString = DateTimeUtils.localDateTimeToString(endTime2);

        FirePointQueryResultBean firePointQueryResult = firePointQueryManager.queryFirePointBelongToUser(sysUserEntity, countyCode, startTimeString, endTimeString, null, null);
        if (firePointQueryResult.isSuccess()) {
            return outputEncapsulationObject(PromptMessageEnum.SUCCESS, firePointQueryResult.getList(), locale);
        } else {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, firePointQueryResult.getErrorMessage(), locale);
        }
    }

    /**
     * 根据火点编码查询火点信息 (Complete)
     */
    @Override
    public String selectFirePointByCode(String fireCode) {
        if (TextUtils.isEmpty(fireCode)) {
            return outputEncapsulationObject(PromptMessageEnum.PROCESS_FAIL, null, locale);
        }
        Specification<FirePointOrdinaryEntity> specification = (root, criteriaQuery, criteriaBuilder) -> {
            //查询条件集合
            List<Predicate> predicateList = new ArrayList<>();
            predicateList.add(criteriaBuilder.equal(root.get("fireCode").as(String.class), fireCode));
            Predicate[] predicateArray = new Predicate[predicateList.size()];
            return criteriaBuilder.and(predicateList.toArray(predicateArray));
        };
        List<FirePointOrdinaryEntity> firePointOrdinaryList = firePointOrdinaryDao.findAll(specification, Sort.by(Sort.Direction.DESC, "satelliteTime"));
        if (firePointOrdinaryList.size() == 1) {
            AllFirePointVo firePointVo = FirePointConvertUtils.convert(firePointOrdinaryList.get(0));
            return outputEncapsulationObject(PromptMessageEnum.SUCCESS, firePointVo, locale);
        }
        return outputEncapsulationObject(PromptMessageEnum.SUCCESS, null, locale);
    }

}
